#!/usr/bin/env python

"""
Copy this python script to your app base directory to use the tcex_local module.

Example Usage:

Install all the required libraries defined in setup.py.  These will be the same packages
with the same version that will get *packaged* with the app.

./app.py --lib

Package the app for installation on the ThreatConnect platform.  Optionally you can pass
the *--collection* flag that will bundle multiple apps into one "App Collection".  The script
will also move the app package to another folder using the *--zip_out* flag.

./app.py --package
./app.py --package --collection
./app.py --package --collection --zip_out /opt/threatconnect/app/bundled/

Validate the application's install.json file.  The validate command is automatically executed
when packaging an app.  A configuration file name can be passed to the script using
the *--config* argument.  By default the script will check the *install.json* file.

./app.py --validate
./app.py --validate --install_json myapp1.install.json

Run the script locally. The tcex_local module will use the tc.json file to generate
the CLI args required by the script.  Typically an app would ship with a tc.json.template
file that provides example CLI args for the app to be run locally.  The config file supports
multiple configuration for different test/use cases with the default case being named "default".
Use the *--test* arg to pass a selected test.

./app.py --run --script <script name>
./app.py --run --script <script name> --test Test1
"""

""" standard """
""" third-party """
""" custom """
from tcex import TcExLocal

tcex_local = TcExLocal()
args = tcex_local.args

if args.lib:
    tcex_local.gen_lib()
elif args.package:
    tcex_local.package()
elif args.run:
    tcex_local.run()
elif args.validate:
    tcex_local.validate(args.install_json)


""" standard """
import json
import os
import uuid

""" third party """
""" custom """

class Resource(object):
    """Common settings for All ThreatConnect API Endpoints"""

    def __init__(self, tcex):
        """Initialize default class values."""
        self._tcex = tcex

        # request
        self._r = tcex.request
        if tcex._args.tc_proxy_tc:
            self._r.proxies = tcex.proxies

        # common
        self._api_branch = None
        self._api_branch_base = None
        self._api_entity = None
        self._api_uri = None
        self._authorization_method = self._tcex.authorization
        self._http_method = 'GET'
        self._filters = []
        self._filter_or = False
        self._pivot = False
        self._parent = None
        self._request_entity = None
        self._request_uri = None
        self._result_count = None
        self._result_limit = 500
        self._result_start = 0
        self._stream = False
        self._status_codes = {}
        self._url = None

    def add_filter(self, name, operator, value):
        """Add ThreatConnect API Filter for this resource request.

        Internal Reference:
            https://docs.google.com/document/d/1mZx9qF9o9RqVGXd_hT6t55aqLtPyzTL2zWRVQsB3HiM

        External Reference:
            https://docs.threatconnect.com

        Args:
            name (string): The filter field name.
            operator (string): The filter comparison operator.
            value (string): The filter value.
        """
        self._filters.append({
            'name': name,
            'operator': operator,
            'value': value
        })

    @property
    def api_branch(self):
        """The ThreatConnect API branch for this resource.

        The **addresses** endpoint from ``/v2/indicators/addresses/``.
        """
        return self._api_branch

    @property
    def api_branch_base(self):
        """The ThreatConnect API branch base (parent branch) for this resource.

        The **indicators** endpoint from ``/v2/indicators`` or ``/v2/indicators/addresses``.
        """
        return self._api_branch_base

    @property
    def api_entity(self):
        """The ThreatConnect API entity for this resource.

        The **address** JSON entity from JSON response to ``/v2/indicators/addresses``.
        """
        return self._api_entity

    @property
    def api_uri(self):
        """The ThreatConnect API URI for this resource.

        The API URI endpoint ``/v2/indicators/addresses``.
        """
        return self._api_uri

    def association_custom(self, resource_api_branch, association_name):
        """Association pivot for this resource with resource value.

        .. Attention:: New untested method

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/indicators/{indicatorType}/{uniqueId}/associations/{associationName}/indicators                 |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{indicatorType}/{uniqueId}/associations/{associationName}/indicators/{indicatorType} |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_api_branch (string): The resource pivot api branch.
            resource_value (string): The resource pivot value (group id or indicator).
            association_name (string): The name of the custom association as defined in the UI.
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = '{}/{}/associations/{}/{}'.format(
            resource_api_branch, association_name, self._request_uri)

    def association_pivot(self, resource_api_branch):
        """Pivot point on association for this resource.

        This method will return all *resources* (group, indicators, task,
        victims, etc) for this resource that are associated with the provided
        resource_type and value.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/groups/{pivot resourceType}/{pivot uniqueId}/{resourceType}                                     |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{pivot resourceType}/{pivot uniqueId}/{resourceType}/{uniqueId}                          |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/groups/{pivot resourceType}/{pivot uniqueId}/{resourceType}/{uniqueId}                          |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{pivot resourceType}/{pivot uniqueId}/{resourceType}                                 |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{pivot resourceType}/{pivot uniqueId}/{resourceType}/{uniqueId}                      |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/indicator/{pivot resourceType}/{pivot uniqueId}/{resourceType}/{uniqueId}                       |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_api_branch (string): The resource pivot api branch including resource id.
            resource_value (string): The resource pivot value (group id or indicator).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = '{}/{}'.format(resource_api_branch, self._request_uri)
        # bcs ^ by using request_uri a user can change pivots without having to worry about the
        #       uri getting mangled, but adding id's and indicators gets overwritten.  Since
        #       the uri should get reset after the request is made is should be okay to
        #       use request_uri here.

    def attributes(self, resource_id=None):
        """Attribute endpoint for this resource with optional attribute id.

        This method will set the resource endpoint for working with Attributes.
        The HTTP GET method will return all attributes applied to this resource
        or if a resource id (attribute id) is provided it will return the
        provided attribute if exists on this resource. An attribute can be added
        to this resource using the HTTP POST method and passing a JSON body
        containing the attribute type and attribute value. Using the HTTP PUT
        method with a provided resource id an attribute can be updated. The
        HTTP DELETE method will remove the provided attribute from this
        resource.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/groups/{resourceType}/{uniqueId}/attributes                                                     |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{resourceType}/{uniqueId}/attributes/{resourceId}                                        |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | DELETE       | /v2/groups/{resourceType}/{uniqueId}/attributes/{resourceId}                                        |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/groups/{resourceType}/{uniqueId}/attributes                                                     |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | PUT          | /v2/groups/{resourceType}/{uniqueId}/attributes/{resourceId}                                        |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (Optional [string]): The resource id (attribute id).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = '{}/attributes'.format(self._request_uri)
        self._request_entity = 'attribute'
        if resource_id is not None:
            self._request_uri = '{}/{}'.format(self._request_uri, resource_id)

    def authorization_method(self, method):
        """Method to create authorization header for this resource request.

        Args:
            method (method): The method to use to genearate the authorization header(s).
        """
        self._authorization_method = method

    @property
    def body(self):
        """The body for this resource request."""
        return self._r.body

    @body.setter
    def body(self, data):
        """The POST/PUT body content for this resource request."""
        self._r.body = data

    @property
    def content_type(self):
        """The Content-Type header value for this resource request."""
        self._r.content_type

    @content_type.setter
    def content_type(self, data):
        """The Content-Type header for this resource request."""
        self._r.content_type = data

    def file_action(self, resource_id, action_name):
        """File action pivot for this resource.

        .. Attention:: New untested method

        .. Note:: Possible action_names are: drop, archive or traffic

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/indicators/files/{uniqueId}/actions/{actionName}/indicators/                                    |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/files/{uniqueId}/actions/{actionName}/indicators/{type}                              |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (string): The resource pivot id (file hash).
            action_name (string): The name of the action as defined by Threatconnect.
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = '{}/{}/actions/{}/{}'.format(
            resource_api_branch, resource_value, association_name, self._request_uri)

    def group_pivot(self, resource_type, resource_id):
        """Pivot point on groups for this resource.

        This method will return all *resources* (indicators, tasks, victims,
        etc) for this resource that are associated with the provided resource
        id (indicator value).

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/groups/{resourceType}/{resourceId}/indicators/{resourceType}                                    |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{resourceType}/{resourceId}/indicators/{resourceType}/{uniqueId}                         |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{resourceType}/{resourceId}/tasks/                                                       |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{resourceType}/{resourceId}/tasks/{uniqueId}                                             |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{resourceType}/{resourceId}/victims/                                                     |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{resourceType}/{resourceId}/victims/{uniqueId}                                           |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_type (string): The resource pivot resource type (indicator type).
            resource_id (integer): The resource pivot id (indicator value).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = 'groups/{}/{}/{}'.format(
            resource_type, resource_id, self._request_uri)

    @property
    def http_method(self):
        """The HTTP Method for this resource request."""
        return self._r.http_method

    @http_method.setter
    def http_method(self, data):
        """The HTTP Method for this resource request."""
        data = data.upper()
        if data in ['DELETE', 'GET', 'POST', 'PUT']:
            self._r.http_method = data
            self._http_method = data

    def indicator_pivot(self, resource_type, resource_id):
        """Pivot point on indicators for this resource.

        This method will return all *resources* (groups, tasks, victims, etc)
        for this resource that are associated with the provided resource id
        (indicator value).

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/indicators/{resourceType}/{resourceId}/groups/{resourceType}                                    |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{resourceType}/{resourceId}/groups/{resourceType}/{uniqueId}                         |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{resourceType}/{resourceId}/tasks/                                                   |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{resourceType}/{resourceId}/tasks/{uniqueId}                                         |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{resourceType}/{resourceId}/victims/                                                 |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{resourceType}/{resourceId}/victims/{uniqueId}                                       |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_type (string): The resource pivot resource type (indicator type).
            resource_id (integer): The resource pivot id (indicator value).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = 'indicators/{}/{}/{}'.format(
            indicator_type, indicator, self._request_uri)

    @property
    def name(self):
        """The name value for this resource.

        The name of the Resource Type (e.g. Indicator, Task, etc.)
        """
        return self._name

    @property
    def owner(self):
        """The Owner payload value for this resource request."""
        return self._r.payload.get('owner')

    @owner.setter
    def owner(self, data):
        """The Owner payload value for this resource request."""
        if data is not None:
            self._r.add_payload('owner', data)
        else:
            self._tcex.log.warn('Provided owner was invalid. ({})'.format(data))

    @property
    def parent(self):
        """The parent object name for this resource."""
        return self._parent

    def paginate(self):
        """Paginate results from ThreatConnect API

        Some API Endpoints require pagination.  This method will paginate through all results and
        return the combined data.

        Return:
            (dictionary): Resource Data
        """
        resources = []
        self._r.add_payload('resultStart', self._result_start)
        self._r.add_payload('resultLimit', self._result_limit)
        results = self.request()
        response = results.get('response')
        if results.get('status') == 'Success':
            data = response.json()['data']
            resources = data[self.request_entity]

            # set results count returned by first API call
            if data.get('resultCount') is not None:
                self._result_count = data.get('resultCount')
                # self._result_start = self._result_limit

            self._tcex.log.debug('Result Count: {}'.format(self._result_count))

            while True:
                if len(resources) >= self._result_count:
                    break
                self._result_start += self._result_limit
                self._r.add_payload('resultStart', self._result_start)
                results = self.request()
                resources.extend(results['data'])

            self._tcex.log.debug('Resource Count: {}'.format(len(resources)))

        return resources

    def request(self):
        """Send the request to the API.

        This method will send the request to the API.  It will try to handle
        all the types of responses and provide the relevant data when possible.
        Some basic error detection and handling is implemented, but not all failure
        cases will get caught.

        Return:
            (dictionary): Response/Results data.
        """
        if self._url is None:
            err = 'URL must not be None Type.'
            self._tcex.log.error(err)
            raise AttributeError(err)

        data = []

        self._r.authorization_method(self._authorization_method)
        self._r.url = '{}/v2/{}'.format(self._url, self._request_uri)
        self._tcex.log.debug('Resource URL: ({})'.format(self._r.url))

        filters = []
        for f in self._filters:
            filters.append('{}{}{}'.format(f['name'], f['operator'], f['value']))
        self._tcex.log.debug('filters: {}'.format(filters))

        if len(filters) > 0:
            self._r.add_payload('filters', ','.join(filters))

        results = {
            'data': [],
            'response': None,
            'result_count': None,
            'status': None
        }

        response = self._r.send(stream=self._stream)
        results['response'] = response
        try:
            if response.status_code in self._status_codes[self._http_method]:
                if response.headers['content-type'] == 'application/json':
                    # handle bulk
                    if self._api_branch == 'bulk':
                        # write bulk download to disk with uniquie ID
                        temp_file = os.path.join(self._tcex._args.tc_temp_path, '{}.json'.format(uuid.uuid4()))
                        self._tcex.log.debug('temp json file: {}'.format(temp_file))
                        with open(temp_file, 'wb') as fh:
                            for block in response.iter_content(1024):
                                fh.write(block)
                        with open(temp_file, 'r') as fh:
                            data = json.load(fh)
                    else:
                        # data = response.json()
                        data = json.loads(response.text)
                        results['status'] = data.get('status')

                    if self._api_branch == 'bulk':
                        # bulk response - {"indicator" : [{........
                        results['data'] = data[self.request_entity]
                    elif 'data' not in data:
                        # bcs - not sure about this one.  may need a better check for results
                        #       that only returns {"status":"Success"}
                        results['data'] = []
                    elif data.get('status') == 'Success':
                        """ this should be the most common outcome """
                        results['data'] = data['data'][self.request_entity]
                    else:
                        self._tcex.message_tc('Failed Request: {}'.format(response.text))
                elif response.headers['content-type'] == 'application/octet-stream':
                    results['data'] = response.content
                else:
                    err = 'Failed Request: {}'.format(response.text)
                    self._tcex.log.error(err)
                    self._tcex.message_tc(err)
            else:
                results['status'] = 'Failure'
                err = 'Failed Request for {}: ({})'.format(self._name, response.text)
                self._tcex.log.error(err)

        except KeyError as e:
            results['status'] = 'Failure'
            msg = 'Error: Invalid key {}. [{}] '.format(e, self.request_entity)
            self._tcex.message_tc(msg)
            self._tcex.log.error(msg)
        except ValueError as e:
            results['status'] = 'Failure'
            msg = 'Error: ({})'.format(e)
            self._tcex.message_tc(msg)
            self._tcex.log.error(msg)

        # bcs - to reset or not to reset?
        self._r.body = None
        # self._r.reset_headers()
        # self._r.reset_payload()
        self._pivot = False
        self._request_uri = self._api_uri
        self._request_entity = self._api_entity
        return results

    @property
    def request_entity(self):
        """The temporary entity name used for the request.

        The request entity starts as the resource api_entity and changes
        depending on the pivot resource. This value is reset after the
        *request()* method is called.
        """
        return self._request_entity

    @property
    def request_uri(self):
        """The temporary uri used for the request.

        The request uri starts as the resource api_uri and changes depending on
        the pivot resource. This value is reset after the *request()* method is
        called.
        """
        return self._request_uri

    def security_label_pivot(self, resource_id):
        """Pivot point on security labels for this resource.

        This method will return all *resources* (group, indicators, task,
        victims, etc) for this resource that have the provided security
        label applied.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/securityLabels/{resourceId}/groups/{resourceType}                                               |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/securityLabels/{resourceId}/groups/{resourceType}/{uniqueId}                                    |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/securityLabels/{resourceId}/indicators/{resourceType}                                           |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/securityLabels/{resourceId}/indicators/{resourceType}/{uniqueId}                                |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (string): The resource pivot id (security label name).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = 'securityLabels/{}/{}'.format(resource_id, self._request_uri)

    def security_labels(self, resource_id=None):
        """Security Label endpoint for this resource with optional label name.

        This method will set the resource endpoint for working with Security
        Labels.  The HTTP GET method will return all security labels applied
        to this resource or if a resource id (security label name) is provided
        it will return the provided security label if it has been applied,
        which could be useful to verify a security label is applied.  The
        provided resource_id (security label name) can be applied to this
        resource using the HTTP POST method.  The HTTP DELETE method will
        remove the provided security label from this resource.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/{resourceType}/{uniqueId}/securityLabels                                                        |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/{resourceType}/{uniqueId}/securityLabels/{resourceId}                                           |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | DELETE       | /v2/{resourceType}/{uniqueId}/securityLabels/{resourceId}                                           |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/{resourceType}/{uniqueId}/securityLabels/{resourceId}                                           |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (Optional [string]): The resource id (security label name).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = '{}/securityLabels'.format(self._request_uri)
        self._request_entity = 'securityLabel'
        if resource_id is not None:
            self._request_uri = '{}/{}'.format(self._request_uri, resource_id)

    def tag_pivot(self, resource_id):
        """Pivot point on tags for this resource.

        This method will return all *resources* (group, indicators, task,
        victims, etc) for this resource that have the provided tag applied.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/tags/{resourceId}/groups/{resourceType}                                                         |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/tags/{resourceId}/groups/{resourceType}/{uniqueId}                                              |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/tags/{resourceId}/indicators/{resourceType}                                                     |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/tags/{resourceId}/indicators/{resourceType}/{uniqueId}                                          |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/tags/{resourceId}/groups/{resourceType}/{uniqueId}                                              |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/tags/{resourceId}/indicators/{resourceType}/{uniqueId}                                          |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (string): The resource pivot id (tag name).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = 'tags/{}/{}'.format(
            self._tcex.safetag(resource_id), self._request_uri)

    def tags(self, resource_id=None):
        """Tag endpoint for this resource with optional tag name.

        This method will set the resource endpoint for working with Tags. The
        HTTP GET method will return all tags applied to this resource or if a
        resource id (tag name) is provided it will return the provided tag if
        it has been applied, which could be useful to verify a tag is applied.
        The provided resource_id (tag) can be applied to this resource using
        the HTTP POST method.  The HTTP DELETE method will remove the provided
        tag from this resource.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/groups/{resourceType}/{uniqueId}/tags                                                           |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/groups/{resourceType}/{uniqueId}/tags/{resourceId}                                              |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{resourceType}/{uniqueId}/tags                                                       |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/indicators/{resourceType}/{uniqueId}/tags/{resourceId}                                          |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | DELETE       | /v2/groups/{resourceType}/{uniqueId}/tags/{resourceId}                                              |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | DELETE       | /v2/indicators/{resourceType}/{uniqueId}/tags/{resourceId}                                          |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/groups/{resourceType}/{uniqueId}/tags/{resourceId}                                              |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | POST         | /v2/indicators/{resourceType}/{uniqueId}/tags/{resourceId}                                          |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (Optional [string]): The resource id (tag name).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = '{}/tags'.format(self._request_uri)
        self._request_entity = 'tag'
        if resource_id is not None:
            self._request_uri = '{}/{}'.format(
                self._request_uri, self._tcex.safetag(resource_id))

    def task_pivot(self, resource_id):
        """Pivot point on Tasks for this resource.

        This method will return all *resources* (group, indicators, victims,
        etc) for this resource that are associated with the provided task id.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/tasks/{resourceId}/groups/{resourceType}                                                        |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/tasks/{resourceId}/groups/{resourceType}/{uniqueId}                                             |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/tasks/{resourceId}/indicators/{resourceType}                                                    |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/tasks/{resourceId}/indicators/{resourceType}/{uniqueId}                                         |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (integer): The resource pivot id (task id).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = 'tasks/{}/{}'.format(
            resource_id, self._request_uri)

    @property
    def value_fields(self):
        """The value fields for this resource.

        The fields in the response JSON that have the key value
        (e.g. ['md5', 'sha1', 'sha256'] or ['ip']).
        """
        return self._value_fields

    def victim_pivot(self, resource_id):
        """Pivot point on Victims for this resource.

        This method will return all *resources* (group, indicators, task,
        etc) for this resource that are associated with the provided victim id.

        **Example Endpoints URI's**

        +--------------+-----------------------------------------------------------------------------------------------------+
        | HTTP Method  | API Endpoint URI's                                                                                  |
        +==============+=====================================================================================================+
        | GET          | /v2/victims/{resourceId}/groups/{resourceType}                                                      |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/victims/{resourceId}/groups/{resourceType}/{uniqueId}                                           |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/victims/{resourceId}/indicators/{resourceType}                                                  |
        +--------------+-----------------------------------------------------------------------------------------------------+
        | GET          | /v2/victims/{resourceId}/indicators/{resourceType}/{uniqueId}                                       |
        +--------------+-----------------------------------------------------------------------------------------------------+

        Args:
            resource_id (integer): The resource pivot id (victim id).
        """
        if self._pivot:
            self._tcex.log.warn('Overriding previous pivot')

        self._pivot = True
        self._request_uri = 'victims/{}/{}'.format(
            resource_id, self._request_uri)

    @property
    def url(self):
        """The ThreatConnect URL for this resource request."""
        return self._url

    @url.setter
    def url(self, data):
        """The ThreatConnect URL for this resource request."""
        self._url = data

    def __str__(self):
        """A printable string for this resource."""
        # TODO: update to include resource specific data.
        printable_string = '\n{0!s:_^80}\n'.format('Resource')

        # body
        printable_string += '\n{0!s:40}\n'.format('Body')
        printable_string += '  {0!s:<29}{1!s:<50}\n'.format('Body', self._r.body)

        # headers
        if len(self._r.headers) > 0:
            printable_string += '\n{0!s:40}\n'.format('Headers')
            for k, v in self._r.headers.items():
                printable_string += '  {0!s:<29}{1!s:<50}\n'.format(k, v)

        # payload
        if len(self._r.payload) > 0:
            printable_string += '\n{0!s:40}\n'.format('Payload')
            for k, v in self._r.payload.items():
                printable_string += '  {0!s:<29}{1!s:<50}\n'.format(k, v)

        return printable_string

#
# Batch
#

class Batch(Resource):
    """Batch Resource Class"""

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Batch, self).__init__(tcex)
        self._api_branch = 'batch'
        self._api_branch_base = self._api_branch # only on parent
        self._api_entity = 'batchId'
        self._api_uri = self._api_branch
        self._name = 'Batch'
        self._parent = 'Batch'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._status_codes = {
            'DELETE': [200],
            'GET': [200],
            'POST': [200, 201],
            'PUT': [200]
        }
        # self._value_fields = ['summary']

    def batch_id(self, batch_id):
        """The ID of the batch job used to push data and/or retrieve status.

        Args:
            batch_id (integer): The id of the batch job.
        """
        self._request_uri = '{}/{}'.format(self._api_uri, batch_id)
        self._request_entity = 'batchStatus'

    def errors(self, batch_id):
        """Update the URI to retrieve errors for a batch job.

        Args:
            batch_id (integer): The id of the batch job.
        """
        self._request_uri = '{}/{}/errors'.format(self._api_uri, batch_id)

#
# Indicator
#

class Indicator(Resource):
    """Indicator Resource Class

    This resource class is the base for all indicators and will return
    indicators of all types.  For specific indicator types use the child
    class of the type required.  Custom indicator types are suppored
    dynamically and are not defined here.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Indicator, self).__init__(tcex)
        self._api_branch = 'indicators'
        self._api_branch_base = self._api_branch # only on parent
        self._api_entity = 'indicator'
        self._api_uri = self._api_branch
        self._name = 'Indicator'
        self._parent = 'Indicator'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._status_codes = {
            'DELETE': [200],
            'GET': [200],
            'POST': [200, 201],
            'PUT': [200]
        }
        self._value_fields = ['summary']

    def indicator(self, data):
        """Update the request URI to include the Indicator for specific indicator retrieval.

        Args:
            data (string): The indicator value
        """
        if self._name != 'Bulk' or self._name != 'Indicator':
            self._request_uri = '{}/{}'.format(self._request_uri, data)

    def resource_id(self, data):
        """Alias for indicator method.

        The resource id for an indicator in this class is the indicator value and not
        the actual indicator id stored in ThreatConnect.

        Args:
            data (string): The indicator value.
        """
        self.indicator(data)


class Address(Indicator):
    """Address Resource Class

    This resource class will return indicators of type Address (ipv4 and/or
    ipv6). To filter on specific indicators use the **indicator** or **resoure_id**
    methods provided in the parent Class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Address, self).__init__(tcex)
        self._api_branch = 'addresses'
        self._api_entity = 'address'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Address'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._value_fields = ['ip']


class Bulk(Indicator):
    """Bulk Resource Class

    This resource class will return bulk status or bulk indicators via the Bulk
    API endpoint.  The base URL will return a status of bulk generation (see
    example below), while the **/csv** and **/json** endpoints will return the
    indicator is the selected format.

    Base response::

        {
            "status": "Success",
            "data": {
                "bulkStatus": {
                    "name": "Acme Corp",
                    "csvEnabled": false,
                    "jsonEnabled": true,
                    "nextRun": "2016-12-07T00:00:00Z",
                    "lastRun": "2016-12-06T00:04:33Z",
                    "status": "Complete"
                }
            }
        }
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Bulk, self).__init__(tcex)
        self._api_branch = 'bulk'
        self._api_entity = 'indicator'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Indicator'  # bcs - should this be bulk?
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._value_fields = ['summary']

    def csv(self, ondemand=False):
        """Update request URI to return CSV data.

        For onDemand bulk generation to work it must first be enabled in the
        ThreatConnect platform under System settings.

        Args:
            ondemand (boolean): Enable on demand bulk generation.
        """
        self._request_uri = '{}/{}'.format(self._api_uri, 'csv')
        self._stream = True
        if ondemand:
            self._r.add_payload('runNow', True)

    def json(self, ondemand=False):
        """Update request URI to return JSON data.

        For onDemand bulk generation to work it must first be enabled in the
        ThreatConnect platform under System settings.

        Args:
            ondemand (boolean): Enable on demand bulk generation.
        """
        self._request_uri = '{}/{}'.format(self._api_uri, 'json')
        self._stream = True
        if ondemand:
            self._r.add_payload('runNow', True)


class EmailAddress(Indicator):
    """EmailAddress Resource Class

    This resource class will return indicators of type Email. To filter on
    specific indicators use the **indicator** or **resoure_id** methods provided
    in the parent Class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(EmailAddress, self).__init__(tcex)
        self._api_branch = 'emailAddresses'
        self._api_entity = 'emailAddress'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'EmailAddress'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._value_fields = ['address']


class File(Indicator):
    """File Resource Class

    This resource class will return indicators of type File (e.g md5, sha1,
    sha256). To filter on specific indicators use the **indicator** or
    **resoure_id** methods provided in the parent Class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(File, self).__init__(tcex)
        self._api_branch = 'files'
        self._api_entity = 'file'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'File'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._value_fields = ['md5', 'sha1', 'sha256']

    def occurrence(self, indicator):
        """Update the URI to retrieve file occurrences for the provided indicator.

        Args:
            indicator (string): The indicator to retrieve file occurrences.
        """
        self._request_entity = 'fileOccurrence'
        self._request_uri = '{}/{}/fileOccurrences'.format(self._api_uri, indicator)


class Host(Indicator):
    """Host Resource Class

    This resource class will return indicators of type Host. To filter on
    specific indicators use the **indicator** or **resoure_id** methods provided
    in the parent Class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Host, self).__init__(tcex)
        self._api_branch = 'hosts'
        self._api_entity = 'host'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Host'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._value_fields = ['hostName']


class URL(Indicator):
    """URL Resource Class

    This resource class will return indicators of type URL. To filter on
    specific indicators use the **indicator** or **resoure_id** methods provided
    in the parent Class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(URL, self).__init__(tcex)
        self._api_branch = 'urls'
        self._api_entity = 'url'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'URL'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._value_fields = ['text']

    def indicator(self, data):
        """Update the request URI to include the Indicator for specific indicator retrieval.

        Overload parent class method to ensure URL indicator is properly encoded.

        Args:
            data (string): The indicator value.
        """
        self._request_uri = '{}/{}'.format(self._api_uri, self._tcex.safeurl(data))

#
# Group
#

class Group(Resource):
    """Group Resource Class

    This resource class is the base for all groups and will return groups of
    all types.  For specific group types use the child class of the type
    required.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Group, self).__init__(tcex)
        self._api_branch = 'groups'
        self._api_branch_base = self._api_branch
        self._api_entity = 'group'
        self._api_uri = self._api_branch
        self._name = 'Group'
        self._parent = 'Group'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._status_codes = {
            'DELETE': [200],
            'GET': [200],
            'POST': [200, 201],
            'PUT': [200, 201]
        }
        self._value_fields = ['name']

    def group_id(self, resource_id):
        """Update the request URI to include the Group ID for specific group retrieval.

        Args:
            resource_id (string): The group id.
        """
        if self._name != 'group':
            self._request_uri = '{}/{}'.format(self._request_uri, resource_id)

    def resource_id(self, resource_id):
        """Alias for group_id method

        Args:
            resource_id (string): The group id.
        """
        self.group_id(resource_id)


class Adversary(Group):
    """Adversary Resource Class

    This resource class will return groups of type Adversary. To filter on
    specific groups use the **group_id** or **resource_id** methods provided in
    the parent class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Adversary, self).__init__(tcex)
        self._api_branch = 'adversaries'
        self._api_entity = 'adversary'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Adversary'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri


class Campaign(Group):
    """Campaign Resource Class

    This resource class will return groups of type Campaign. To filter on
    specific groups use the **group_id** or **resource_id** methods provided in
    the parent class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Campaign, self).__init__(tcex)
        self._api_branch = 'campaigns'
        self._api_entity = 'campaign'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Campaign'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri


class Document(Group):
    """Document Resource Class

    This resource class will return groups of type Document. To filter on
    specific groups use the **group_id** or **resource_id** methods provided in
    the parent class.
    """

    def __init__(self, tcex):
        """Initialize default class values.

        Args:
            tcex (instance): Instance of TcEx Class.
        """
        super(Document, self).__init__(tcex)
        self._api_branch = 'documents'
        self._api_entity = 'document'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Document'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri

    def download(self, resource_id):
        """Update the request URI to download the document for this resource.

        Args:
            resource_id (integer): The group id.
        """
        self.resource_id(resource_id)
        self._request_uri = '{}/download'.format(self._request_uri)

    def upload(self, resource_id, data):
        """Update the request URI to upload the a document to this resource.

        Args:
            resource_id (integer): The group id.
            data (any): The raw data to upload.
        """
        self.body = data
        self.content_type = 'application/octet-stream'
        self.resource_id(resource_id)
        self._request_uri = '{}/upload'.format(self._request_uri)


class Email(Group):
    """Email Resource Class

    This resource class will return groups of type Email. To filter on
    specific groups use the **group_id** or **resource_id** methods provided in
    the parent class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Email, self).__init__(tcex)
        self._api_branch = 'emails'
        self._api_entity = 'email'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Email'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri


class Incident(Group):
    """Incident Resource Class

    This resource class will return groups of type Incident. To filter on
    specific groups use the **group_id** or **resource_id** methods provided in
    the parent class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Incident, self).__init__(tcex)
        self._api_branch = 'incidents'
        self._api_entity = 'incident'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Incident'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri

    def event_date(self, date):
        """Incident Event Date.

        .. Attention:: Not implemented at this time

        Args:
            date: The event date in ISO 8601 format.
        """
        pass


class Signature(Group):
    """Signature Resource Class

    This resource class will return groups of type Signature. To filter on
    specific groups use the **group_id** or **resource_id** methods provided in
    the parent class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Signature, self).__init__(tcex)
        self._api_branch = 'signatures'
        self._api_entity = 'signature'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Signature'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri


class Threat(Group):
    """Threat Resource Class

    This resource class will return groups of type Threat. To filter on
    specific groups use the **group_id** or **resource_id** methods provided in
    the parent class.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Threat, self).__init__(tcex)
        self._api_branch = 'threats'
        self._api_entity = 'threat'
        self._api_uri = '{}/{}'.format(self._api_branch_base, self._api_branch)
        self._name = 'Threat'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
#
# Owner
#

class Owner(Resource):
    """Owner Class

    This resource class will return Owners.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Owner, self).__init__(tcex)
        self._api_branch = 'owners'
        self._api_entity = 'owner'
        self._api_uri = self._api_branch
        self._name = 'Owner'
        self._parent = 'Owner'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._status_codes = {
            'GET': [200]
        }
        self._value_fields = ['name']

    def owner_id(self, resource_id):
        """Update the request URI to include the Owner Id for specific retrieval.

        Args:
            resource_id (string): The owner id.
        """
        self._request_uri = '{}/{}'.format(self._request_uri, resource_id)

    def resource_id(self, resource_id):
        """Alias for owner_id method

        Args:
            resource_id (string): The owner id.
        """
        self.owner_id(resource_id)

#
# Security Label
#

class SecurityLabel(Resource):
    """Security Label Class

    This resource class will return Security Labels.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Task, self).__init__(tcex)
        self._api_branch = 'securityLabels'
        self._api_entity = 'securityLabel'
        self._api_uri = self._api_branch
        self._name = 'SecurityLabel'
        self._parent = 'SecurityLabel'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._status_codes = {
            'DELETE': [200],
            'GET': [200],
            'POST': [200],
            'PUT': [200]
        }
        self._value_fields = ['name']

    def label(self, resource_id):
        """Update the request URI to include the Security Label for specific retrieval.

        Args:
            resource_id (string): The security label.
        """
        self._request_uri = '{}/{}'.format(self._request_uri, resource_id)

    def resource_id(self, resource_id):
        """Alias for label method

        The resource id is the security label name.

        Args:
            resource_id (string): The security label.
        """
        self._request_uri = '{}/{}'.format(self._request_uri, resource_id)

#
# Task
#

class Task(Resource):
    """Task Class

    This resource class will return Tasks.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Task, self).__init__(tcex)
        self._api_branch = 'tasks'
        self._api_entity = 'task'
        self._api_uri = self._api_branch
        self._name = 'Task'
        self._parent = 'Task'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._status_codes = {
            'DELETE': [200],
            'GET': [200],
            'POST': [200],
            'PUT': [200]
        }
        self._value_fields = ['name']

    def resource_id(self, data):
        """Update the request URI to include the Task Id for specific retrieval.

        Args:
            resource_id (string): The task id.
        """
        self._request_uri = '{}/{}'.format(self._request_uri, data)

#
# Victim
#

class Victim(Resource):
    """Victim Class

    This resource class will return Victims.
    """

    def __init__(self, tcex):
        """Initialize default class values."""
        super(Victim, self).__init__(tcex)
        self._api_branch = 'victims'
        self._api_entity = 'victim'
        self._api_uri = self._api_branch
        self._name = 'Victim'
        self._parent = 'Victim'
        self._request_entity = self._api_entity
        self._request_uri = self._api_uri
        self._status_codes = {
            'DELETE': [200],
            'GET': [200],
            'POST': [200],
            'PUT': [200]
        }
        self._value_fields = ['name']

    def resource_id(self, data):
        """Update the request URI to include the Victim Id for specific retrieval.

        Args:
            resource_id (string): The victim id.
        """
        self._request_uri = '{}/{}'.format(self._request_uri, data)

#
# Class Factory
#

def ClassFactory(name, base_class, class_dict):
    """Internal method for dynamically building Custom Indicator classes."""
    def __init__(self, tcex):
        base_class.__init__(self, tcex)
        for k, v in class_dict.items():
            setattr(self, k, v)

    newclass = type(name, (base_class,),{"__init__": __init__})
    return newclass


""" standard """
import time
from json import dumps
from logging import FileHandler, makeLogRecord

""" third-party """
""" custom """


def create_log_entry(record):
    """Create the API log entry.

    Args:
        record (object): The log entry record.

    Returns:
        (dictionary): The data to log to API
    """
    log_entry = {}

    if hasattr(record, 'created'):
        log_entry['timestamp'] = int(float(record.created) * 1000)

    if hasattr(record, 'msg'):
        log_entry['message'] = record.msg

    if hasattr(record, 'levelname'):
        log_entry['level'] = record.levelname

    return log_entry


class ApiLoggingHandler(FileHandler):
    """Extension of FileHandler

    Sends logs entries to the ThreatConnect API.
    """

    def __init__(self, filename, tcex, max_entries_before_flush=100):
        super(ApiLoggingHandler, self).__init__(filename)
        self._tcex = tcex
        self.max_entries_before_flush = max_entries_before_flush

        # init entries
        self.entries = []

    def emit(self, record):
        """Overload of logger emit method

        Args:
            record (object): The log entry record
        """
        entry = create_log_entry(record)
        self.entries.append(entry)

        # if we've reached the max_entries threshold, flush the handler
        # bcs - would doing a json.dumps and checking size for POST be safer
        #       than an arbitrary amount of events?
        if len(self.entries) > self.max_entries_before_flush:
            self.log_to_api()
            self.entries = []

        super(ApiLoggingHandler, self).emit(record)

    def log_to_api(self):
        """Best effort API logger.

        Send logs to API and do nothing if the attempt fails.

        Example log event::

            [{
                "timestamp": 1478907537000,
                "message": "Test Message",
                "level": "DEBUG"
            }]

        """
        if len(self.entries) > 0:
            # Make API call
            r = self._tcex.request
            r.authorization_method(self._tcex.authorization)
            # bcs - sort entry by *created*?
            # r.body = dumps(self.entries)
            r.body = dumps(self.entries, ensure_ascii=False)
            self.entries = []  # clear entries
            r.http_method = 'POST'
            if self._tcex._args.tc_proxy_tc:
                r.proxies = self._tcex._proxy_config
            r.url = '{}/v2/logs/app'.format(self._tcex._args.tc_api_path)
            try:
                results = r.send()
                # if results.headers.get('content-type') == 'application/json':
                #     data = results.json()
                #     if data.get('status') == 'Success':
                #         pass
            except:
                # best effort for now.  don't fret if logging fails
                pass


""" standard """
import socket
import time

""" third-party """
from requests import (exceptions, packages, Request, Session)
# disable ssl warning message
packages.urllib3.disable_warnings()

""" custom """

class TcExRequest(object):
    """ """
    def __init__(self, tcex):
        """ """
        self._tcex = tcex

        self._authorization_method = None
        self._basic_auth = None
        self._body = None
        self._content_type = None
        self._headers = {}
        self._http_method = 'GET'
        self._payload = {}
        self._proxies = {}
        self._url = None
        self.user_agent = 'ThreatConnect'

        # request properties
        self._retries = 4
        self._sleep = 5
        self._session = Session()
        self._timeout = 30
        self._verify_ssl = False

    #
    # Body
    #

    @property
    def body(self):
        """The POST/PUT body content for this request."""
        return self._body

    @body.setter
    def body(self, data):
        """The POST/PUT body content for this request."""
        if data is not None:
            self._body = data
            self.add_header('Content-Length', len(self._body))

    #
    # HTTP Headers
    #

    @property
    def headers(self):
        """The header values for this request."""
        return self._headers

    def reset_headers(self):
        """Reset header dictionary for this request."""
        self._headers = {}

    def add_header(self, key, val):
        """Add a key value pair to header.

        Args:
            key (string): The header key
            value (string): The header value
        """
        self._headers[key] = str(self._tcex.uni(val))

    @property
    def authorization(self):
        """The "Authorization" header value for this request."""
        return self._headers.get('Authorization')

    @authorization.setter
    def authorization(self, data):
        """The "Authorization" header value for this request."""
        self.add_header('Authorization', data)

    def authorization_method(self, method):
        """Method to create Authorization header for this request.

        Args:
            method (method): The method to use to genearate the authorization header(s).
        """
        self._authorization_method = method

    @property
    def basic_auth(self):
        """The basic auth settings for this request."""
        return self._basic_auth

    @basic_auth.setter
    def basic_auth(self, data):
        """The basic auth settings for this request."""
        self._basic_auth = data

    @property
    def content_type(self):
        """The Content-Type header value for this request."""
        return self._content_type

    @content_type.setter
    def content_type(self, data):
        """The Content-Type header value for this request."""
        self._content_type = self._tcex.uni(data)
        self.add_header('Content-Type', data)

    @property
    def user_agent(self):
        """The the User-Agent header value for this request."""
        return self._user_agent

    @user_agent.setter
    def user_agent(self, data):
        """The the User-Agent header value for this request."""
        self.add_header('User-agent', data)

    #
    # HTTP Payload
    #

    @property
    def payload(self):
        """The payload values for this request."""
        return self._payload

    def reset_payload(self):
        """Reset payload dictionary"""
        self._payload = {}

    def add_payload(self, key, val):
        """Add a key value pair to payload for this request.

        Args:
            key (string): The payload key
            value (string): The payload value
        """
        self._payload[key] = str(self._tcex.uni(val))

    #
    # HTTP Method
    #

    @property
    def http_method(self):
        """The HTTP method for this request."""
        return self._http_method

    @http_method.setter
    def http_method(self, data):
        """The HTTP method for this request."""
        data = data.upper()
        if data in ['DELETE', 'GET', 'POST', 'PUT']:
            self._http_method = data

            # set content type for commit methods (best guess)
            if self._content_type is None and data in ['POST', 'PUT']:
                self.add_header('Content-Type', 'application/json')
        else:
            raise AttributeError('Request Object Error: {0!s} is not a valid HTTP method.'.format(data))

    #
    # Send Properties
    #

    @property
    def proxies(self):
        """The proxy settings for this request."""
        return self._proxies

    @proxies.setter
    def proxies(self, data):
        """The proxy settings for this request."""
        self._proxies = data

    @property
    def timeout(self):
        """The HTTP timeout value for this request."""
        return self._timeout

    @timeout.setter
    def timeout(self, data):
        """The HTTP timeout value for this request."""
        if isinstance(data, int):
            self._timeout = data

    @property
    def verify_ssl(self):
        """The SSL validation setting for this request."""
        return self._verify_ssl

    @verify_ssl.setter
    def verify_ssl(self, data):
        """The SSL validation setting for this request."""
        if isinstance(data, bool):
            self._verify_ssl = data

    #
    # Send
    #

    def send(self, stream=False):
        """Send the HTTP request via Python Requests modules.

        This method will send the request to the remote endpoint.  It will try to handle
        temporary communications issues by retrying the request automatically.

        Args:
            stream (boolean): Boolean to enable stream download.

        Returns:
            (Requests.Response) The Request response
        """
        #
        # prepare request
        #

        api_request = Request(
            method=self._http_method, url=self._url, data=self._body, params=self._payload)

        request_prepped = api_request.prepare()

        # add authorization header returned by authorization method
        if self._authorization_method is not None:
            self._headers.update(self._authorization_method(request_prepped))
        request_prepped.prepare_headers(self._headers)
        # self._tcex.log.debug('Request URL: {}'.format(self._url))

        if self._basic_auth is not None:
            request_prepped.prepare_auth(self._basic_auth)

        #
        # api request (gracefully handle temporary communications issues with the API)
        #
        for i in range(1, self._retries + 1, 1):
            try:
                response = self._session.send(
                    request_prepped, proxies=self._proxies, timeout=self._timeout,
                    verify=self._verify_ssl, stream=stream)
                break
            except exceptions.ReadTimeout as e:
                self._tcex.log.error('Error: {0!s}'.format(e))
                self._tcex.log.error('The server may be experiencing delays at the moment.')
                self._tcex.log.info('Pausing for {0!s} seconds to give server time to catch up.'.format(self._sleep))
                time.sleep(self._sleep)
                self._tcex.log.info('Retry {0!s} ....'.format(i))

                if i == self._retries:
                    self._tcex.log.critical('Exiting: {0!s}'.format(e))
                    raise RuntimeError(e)
            except exceptions.ConnectionError as e:
                self._tcex.log.error('Error: {0!s}'.format(e))
                self._tcex.log.error('Connection Error. The server may be down.')
                self._tcex.log.info('Pausing for {0!s} seconds to give server time to catch up.'.format(self._sleep))
                time.sleep(self._sleep)
                self._tcex.log.info('Retry {0!s} ....'.format(i))
                if i == self._retries:
                    self._tcex.log.critical('Exiting: {0!s}'.format(e))
                    raise RuntimeError(e)
            except socket.error as e:
                self._tcex.log.critical('Exiting: {0!s}'.format(e))
                raise RuntimeError(e)

        self._tcex.log.info('URL: {}'.format(response.url))
        self._tcex.log.info('Status Code: {0!s}'.format(response.status_code))
        return response

    #
    # URL
    #

    @property
    def url(self):
        """The URL for this request."""
        return self._url

    @url.setter
    def url(self, data):
        """The URL for this request."""
        self._url = data

    def __str__(self):
        """Print this request instance configuration."""
        printable_string = '\n{0!s:_^80}\n'.format('Request')

        #
        # http settings
        #
        printable_string += '\n{0!s:40}\n'.format('HTTP Settings')
        printable_string += '  {0!s:<29}{1!s:<50}\n'.format('HTTP Method', self.http_method)
        printable_string += '  {0!s:<29}{1!s:<50}\n'.format('Request URL', self.url)
        printable_string += '  {0!s:<29}{1!s:<50}\n'.format('Content Type', self.content_type)
        printable_string += '  {0!s:<29}{1!s:<50}\n'.format('Body', self.body)

        #
        # headers
        #
        if len(self.headers) > 0:
            printable_string += '\n{0!s:40}\n'.format('Headers')
            for k, v in self.headers.items():
                printable_string += '  {0!s:<29}{1!s:<50}\n'.format(k, v)

        #
        # payload
        #
        if len(self.payload) > 0:
            printable_string += '\n{0!s:40}\n'.format('Payload')
            for k, v in self.payload.items():
                printable_string += '  {0!s:<29}{1!s:<50}\n'.format(k, v)

        return printable_string


""" standard """
import json
import re
import time

""" third-party """

""" custom """
import tcex_resources as Resources


class TcExJob(object):
    """Job processing functionality

    Supports batch indicator adds and allows structured group adds.
    """

    def __init__(self, tcex):
        """Initialize class data

        Args:
            tcex (instance): An instance of TcEx class
        """
        self._tcex = tcex

        # containers
        self._file_occurrences = []
        self._group_cache = {}
        self._group_associations = []
        self._groups = []
        self._indicators = []

        # batch "bad" errors
        self._batch_failures = [
            'would exceed the number of allowed indicators'
        ]

        # keep the batch_id
        self._indicator_batch_ids = []

    def _chunk_indicators(self):
        """Split indicator list into smaller more manageable numbers.

        """
        for i in xrange(0, len(self._indicators), self._tcex._args.batch_chunk):
            yield self._indicators[i:i+self._tcex._args.batch_chunk]

    def _group_add(self, resource_type, resource_name, owner, data=None):
        """Add a group to ThreatConnect

        Args:
            resource_type (string): String with resource [group] type
            resource_name (string): The name of the group
            owner (string): The name of the TC owner for the group add
            data (Optional [dictionary]): Any optional group parameters, tags,
                or attributes

        Returns:
            (integer): The ID of the created resource
        """
        resource_id = None
        if data is None:
            data = {}

        self._tcex.log.debug(u'Adding {} "{}" in owner {}'.format(
            resource_type, resource_name, owner))
        resource_body = {'name': resource_name}
        resource = getattr(Resources, resource_type)(self._tcex)
        resource.http_method = 'POST'
        resource.owner = owner
        resource.url = self._tcex._args.tc_api_path

        # incident
        if data.get('eventDate') is not None:
            resource_body['eventDate'] = data.get('eventDate')

        resource.body = json.dumps(resource_body)
        results = resource.request()

        if results.get('status') == 'Success':
            resource_id = results['data']['id']
            resource = getattr(Resources, resource_type)(self._tcex)
            resource.http_method = 'POST'
            resource.url = self._tcex._args.tc_api_path

            # add attributes
            for attribute in data.get('attributes', []):
                self._tcex.log.debug('Adding attribute type ({})'.format(attribute['type']))
                resource.resource_id(resource_id)
                resource.attributes()
                resource.body = json.dumps(attribute)
                results = resource.request()

            # add tags
            for tag in data.get('tags', []):
                self._tcex.log.debug(u'Adding tag ({})'.format(tag))
                resource.resource_id(resource_id)
                resource.tags(tag)
                results = resource.request()
        else:
            err = 'Failed adding group ({})'.format(resource_name)
            self._tcex.log.error(err)
            self._tcex.exit_code(3)

        return resource_id

    def _process_file_occurrences(self, owner):
        """Process file occurrences and write to TC API

        Args:
            owner (string):  The owner name for the indicator to be written
        """
        # POST /v2/indicators/files/BE7DE2F0CF48294400C714C9E28ECD01/fileOccurrences
        # supported format -> 2014-11-03T00:00:00-05:00

        resource = getattr(Resources, 'File')(self._tcex)
        resource.http_method = 'POST'
        resource.owner = owner
        resource.url = self._tcex._args.tc_api_path

        for occurrence in self._file_occurrences:
            # remove the hash from the dictionary and add to URI
            resource.occurrence(occurrence.pop('hash'))
            resource.body = json.dumps(occurrence)
            results = resource.request()

    def _process_group_association(self, owner):
        """Process groups and write to TC API

        Args:
            owner (string):  The owner name for the indicator to be written
        """
        for ga in self._group_associations:
            self._tcex.log.info(u'creating association: group {} indicator {}'.format(
                    ga['group_name'], ga['indicator']))

            group_id = self.group_id(ga['group_name'], owner, ga['group_type'])
            if group_id is None:
                continue

            resource = getattr(Resources, ga['indicator_type'])(self._tcex)
            resource.http_method = 'POST'
            resource.indicator(ga['indicator'])
            resource.owner = owner
            resource.url = self._tcex._args.tc_api_path

            association_resource = getattr(Resources, ga['group_type'])(self._tcex)
            association_resource.resource_id(group_id)
            resource.association_pivot(association_resource.request_uri)
            resource.request()

    def _process_groups(self, owner, duplicates=False):
        """Process groups and write to TC API

        Args:
            owner (string):  The owner name for the indicator to be written
        """
        for group in self._groups:
            cache = self.group_cache(owner, group['type'])

            self._tcex.log.debug(u'Processing group ({})'.format(group.get('name')))
            if duplicates:
                # bcs - support duplicate groups
                pass
            elif group['name'] not in cache.keys():
                group_id = self._group_add(group['type'], group['name'], owner, group)
                if group_id is not None:
                    self._group_cache[owner][group['type']][group['name']] = group_id
                self._tcex.log.info(u'Creating group {} [{}]'.format(group['name'], group_id))
            else:
                self._tcex.log.debug(u'Existing Group ({})'.format(group.get('name')))

    def _process_indicators(self, owner):
        """Process batch indicators and write to TC API

        Args:
            owner (string):  The owner name for the indicator to be written
        """
        self._tcex.log.info('Processing {} indicators'.format(len(self._indicators)))

        resource = getattr(Resources, 'Batch')(self._tcex)
        resource.http_method = 'POST'
        resource.url = self._tcex._args.tc_api_path

        batch_job_body = {
            'action': self._tcex._args.batch_action,
            'attributeWriteType': self._tcex._args.batch_write_type,
            'haltOnError': self._tcex._args.batch_halt_on_error,
            'owner': owner
        }

        for chunk in self._chunk_indicators():
            self._tcex.log.info('Batch Chunk Size: {}'.format(len(chunk)))
            resource.content_type = 'application/json'
            resource.body = json.dumps(batch_job_body)
            results = resource.request()

            if results['status'] == 'Success':
                batch_id = results['data']
                self._indicator_batch_ids.append(batch_id)

                resource.content_type = 'application/octet-stream'
                resource.batch_id(batch_id)
                resource.body = json.dumps(chunk)
                results = resource.request()

                # bcs - add a small delay before first status check then normal delay in loop
                time.sleep(3)

                poll_time = 0
                while True:
                    self._tcex.log.debug('poll_time: {0}'.format(poll_time))
                    if poll_time >= self._tcex._args.batch_poll_interval_max:
                        msg = 'Status check exceeded max poll time.'
                        self._tcex.log.error(msg)
                        self._tcex.message_tc(msg)
                        self._tcex.exit(1)

                    if self.batch_status(batch_id):
                        break

                    time.sleep(self._tcex._args.batch_poll_interval)

                    poll_time += self._tcex._args.batch_poll_interval

    def batch_action(self, action):
        """Set the default batch action for argument parser.

        Args:
            action (string): Set batch job action
        """
        if action in ['Create', 'Delete']:
            self._tcex._parser.set_defaults(batch_action=action)

    def batch_chunk(self, chunk_size):
        """Set batch chunk_size for argument parser.

        Args:
            chunk_size (integer): Set batch job chunk size
        """
        self._tcex._parser.set_defaults(batch_chunk=chunk_size)

    def batch_halt_on_error(self, halt_on_error):
        """Set batch halt on error boolean for argument parser.

        Args:
            halt_on_error (boolean): Boolean value for halt on error
        """
        if isinstance(halt_on_error, bool):
            self._tcex._parser.set_defaults(batch_halt_on_error=halt_on_error)

    def batch_indicator_success(self):
        """Check completion for all batch jobs associated with this instance.

        Iterate over self._indicator_batch_ids set in
        :py:meth:`~tcex.tcex_job.TcExJob._process_indicators` and return the status of all job.

        Returns:
            (dictionary): The status results from the Batch jobs.
        """

        status = {'success': 0, 'failure': 0, 'unprocessed': 0}

        resource = getattr(Resources, 'Batch')(self._tcex)
        resource.url = self._tcex._args.tc_api_path

        for batch_id in self._indicator_batch_ids:
            resource.batch_id(batch_id)
            results = resource.request()

            if (results.get('status') == 'Success' and
                'batchStatus' in results['data'] and
                results['data']['batchStatus'] == 'Completed'):

                status['success'] += results['data']['batchStatus']['successCount']
                status['failure'] += results['data']['batchStatus']['errorCount']
                status['unprocessed'] += results['data']['batchStatus']['unprocessCount']

        return status

    def batch_poll_interval(self, interval):
        """Set batch polling interval for argument parser.

        Args:
            interval (integer): Seconds between polling
        """
        if isinstance(interval, int):
            self._tcex._parser.set_defaults(batch_poll_interval=interval)

    def batch_poll_interval_max(self, interval_max):
        """Set batch polling interval max for argument parser.

        Args:
            interval_max (integer): Max seconds before timeout on batch
        """
        if isinstance(interval_max, int):
            self._tcex._parser.set_defaults(batch_poll_interval_max=interval_max)

    def batch_status(self, batch_id):
        """Check the status of a batch job

        This method will get the status of a batch job. Any errors returned from the batch status
        will be automatically logged.

        Critical errors are defined in the __init__ method in the self._batch_failures dictionary.
        If any of the critical errors are found the execution of the App will halt with a exit
        code of 1.  All other errors will set the status code to 3, but will not cause the execution
        to halt.

        Args:
            batch_id (int): Id of the batch job

        Returns:
            (boolean): Status of batch job.
        """
        status = False

        resource = getattr(Resources, 'Batch')(self._tcex)
        resource.url = self._tcex._args.tc_api_path
        resource.batch_id(batch_id)
        results = resource.request()

        self._tcex.log.info('batch id: {}'.format(batch_id))
        if results.get('status') == 'Success':
            if results['data']['status'] == 'Completed':
                status = True

                status_msg = 'Batch Job completed, totals: '
                status_msg += 'succeeded: {0}, '.format(results['data']['successCount'])
                status_msg += 'failed: {0}, '.format(results['data']['errorCount'])
                status_msg += 'unprocessed: {0}'.format(results['data']['unprocessCount'])
                self._tcex.log.info(status_msg)

                if results['data']['errorCount'] > 0:
                    self._tcex.exit_code(3)

                    resource = getattr(Resources, 'Batch')(self._tcex)
                    resource.url = self._tcex._args.tc_api_path
                    resource.errors(batch_id)
                    error_results = resource.request()

                    for error in json.loads(error_results['data']):
                        error_reason = error.get('errorReason')
                        for error_msg in self._batch_failures:

                            if re.findall(error_msg, error_reason):
                                # Critical Error
                                err = 'Batch Error: {0}'.format(error)
                                self._tcex.message_tc(err)
                                self._tcex.log.error(err)
                                self._tcex.exit(1)
                        self._tcex.log.warn('Batch Error: {0}'.format(error))
            else:
                self._tcex.log.debug('Batch Status: {}'.format(results['data']['status']))

        return status

    def batch_write_type(self, write_type):
        """Set batch attributes write type for argument parser.

        Args:
            write_type (string): Type of Append or Replace
        """
        if write_type in ['Append', 'Replace']:
            self._tcex._parser.set_defaults(batch_write_type=write_type)

    def file_occurrence(self, fo):
        """

        Args:
            fo (dictionary): The file occurrence data.

        .. Warning:: There is no validation of the data passed to this method.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 3

            {
                "date" : "2014-11-03T00:00:00-05:00",
                "fileName" : "win999301.dll",
                "hash": "BE7DE2F0CF48294400C714C9E28ECD01",
                "path" : "C:\\Windows\\System"
            }

        .. Note:: The hash in the example above is not posted in the body, but extracted to use
                  in the URI.
        """
        # POST /v2/indicators/files/BE7DE2F0CF48294400C714C9E28ECD01/fileOccurrences

        if isinstance(fo, list):
            self._file_occurrences.extend(fo)
        elif isinstance(fo, dict):
            self._file_occurrences.append(fo)

    ## bcs - not rewriting this until I know what it is for.
    ## def get_batch_job(self, batch_id):
    ##     """what is this for?"""
    ##
    ##     batch_jobs = tc.batch_jobs()
    ##     batch_filter = batch_jobs.add_filter()
    ##     batch_filter.add_id(batch_id)

    ##     try:
    ##         batch_jobs.retrieve()
    ##     except RuntimeError as re:
    ##         self._tcex.log.error(re)
    ##         self._tcex.message_tc(re)

    ##     return batch_jobs

    def group(self, group):
        """Add group data to TcEx job.

        This method will accept a list or dictionary of formatted *Group* data and submit it to the
        API when the :py:meth:`~tcex.tcex_job.TcExJob.process` method is called.

        .. Warning:: There is no validation of the data passed to this method.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 9,14

            {
              'attribute': [
                {
                  'type': 'Description',
                  'value': 'Test Description'
                }
              ],
              'eventDate': '2015-03-7T00:00:00Z'
              'name': 'adversary-001',
              'tag': [
                'APT',
                'Crimeware'
              ],
              'type': 'Adversary'
            }

        Args:
            group (dict | list): Dictionary or List containing group data.
        """
        if isinstance(group, list):
            self._groups.extend(group)
        elif isinstance(group, dict):
            self._groups.append(group)

    def group_association(self, associations):
        """Add group association data to TcEx job.

        This method will add group association data to the group association list.

        .. Warning:: There is no validation of the data passed to this method.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 2-5

            {
              'group_name': 'adversary-001',
              'group_type': 'Adversary',
              'indicator': '1.1.1.1',
              'indicator_type': 'Address'
            }

        Args:
            associations (dict | list): Dictionary or List containing group
                association data
        """
        if isinstance(associations, list):
            self._group_associations.extend(associations)
        elif isinstance(associations, dict):
            self._group_associations.append(associations)

    @property
    def group_association_len(self):
        """The current length of the group association list.

        Returns:
            (integer): The length of the group association list
        """
        return len(self._group_associations)

    def group_cache(self, owner, resource_type):
        """Cache group data from the ThreatConnect Platform.

        The method will cache ThreatConnect group data by owner and type.

        Args:
            owner (string): The name of the TC owner
            resource (dictionary): Dictonary containing resource parameters

        Returns:
            (dictionary): Dictionary of group resources
        """

        # already cached
        if owner in self._group_cache:
            if resource_type in self._group_cache[owner]:
                return self._group_cache[owner][resource_type]

        self._tcex.log.info(u'Caching group type {} for owner {}'.format(resource_type, owner))
        self._group_cache.setdefault(owner, {})
        self._group_cache[owner].setdefault(resource_type, {})

        resource = getattr(Resources, resource_type)(self._tcex)
        resource.owner = owner
        resource.url = self._tcex._args.tc_api_path
        results = resource.paginate()

        for group in results:
            self._tcex.log.debug('cache - group name: ({})'.format(group.get('name')))
            self._group_cache[owner][resource_type][group['name']] = group['id']

        return self._group_cache[owner][resource_type]

    def group_id(self, name, owner, resource_type):
        """Get the group id from the group cache.

        Args:
            name(string): The name of the Group
            owner (string): The TC Owner where the resouce should be found
            resource (dictionary): Dictionary of resource parameters

        Returns:
            (integer): The ID for the provided group name and owner.
        """
        group_id = None

        if owner not in self._group_cache:
            self.group_cache(owner, resource_type)
        elif resource_type not in self._group_cache[owner]:
            self.group_cache(owner, resource_type)

        if name in self._group_cache[owner][resource_type]:
            group_id = self._group_cache[owner][resource_type][name]

        return group_id

    @property
    def group_len(self):
        """The current length of the group list.

        Returns:
            (integer): The length of the group list.
        """
        return len(self._groups)

    def indicator(self, indicator):
        """Add indicator data to TcEx job.

        This method will add indicator data to this TcEx job to be submitted via batch import.

        .. Warning:: There is no validation of the data passed to this method. Duplicate prevention
                     will only be handled when a dictionary is provided.

        **Example Data** *(required fields are highlighted)*

        .. code-block:: javascript
            :linenos:
            :lineno-start: 1
            :emphasize-lines: 14,19

            {
              'associatedGroup': [
                '1',
                '8'
              ],
              'attribute': [
                {
                  'type': 'Description',
                  'value': 'Test Description'
                }
              ],
              'confidence': 5,
              'rating': '3',
              'summary': '1.1.1.1',
              'tag': [
                'APT',
                'Crimeware'
              ],
              'type': 'Address'
            }

        Args:
            indicator (dict | list): Dictionary or List containing indicator
                data
        """
        if isinstance(indicator, list):
            self._indicators.extend(indicator)
        elif isinstance(indicator, dict):
            # verifiy indicator is not a duplicate before adding
            if indicator['summary'] not in self._indicators:
                self._indicators.append(indicator)

    @property
    def indicator_data(self):
        """Return the current indicator list.

        Returns:
            (list): The indicator list
        """
        return self._indicators

    @property
    def indicator_len(self):
        """The current length of the indicator list

        Returns:
            (integer): The length of the indicator list
        """
        return len(self._indicators)

    def process(self, owner):
        """Process all association, group and indicator data.

        Process each of the supported data types for this job.

        Args:
            owner (string): The owner name for the data to be written
        """
        if len(self._groups) > 0:
            self._tcex.log.info('Processing Groups')
            self._process_groups(owner)

        if len(self._indicators) > 0:
            self._tcex.log.info('Processing Indicators')
            self._process_indicators(owner)

        if len(self._file_occurrences) > 0:
            self._tcex.log.info('Processing File Occurrences')
            self._process_file_occurrences(owner)

        if len(self._group_associations) > 0:
            self._tcex.log.info('Processing Group Associations')
            self._process_group_association(owner)


import os
import sys
import inspect

__author__ = 'ThreatConnect (support@threatconnect.com)'
__version__ = '0.0.11'
__license__ = 'Apache License, Version 2'

## BCS - This is invalid after lib change for multiple Python versions.
##       Not sure if it is required now.

## # For Apps Append the '/lib' directory
## lib_path = os.path.join(os.getcwd(), 'lib')
##
## for root, dirs, files in os.walk(lib_path):
##     while len(dirs) > 0:
##         sys.path.insert(1, '%s/%s' % (lib_path, dirs.pop()))

try:
    from tcex import TcEx
except ImportError as e:
    print('Error: {}'.format(e))
    print('Try app.py --lib')

from tcex_local import TcExLocal
from argparser import ArgParser



""" standard """
import argparse
import json
import os
import re
import shutil
import sys
import time
import zipfile
from setuptools.command import easy_install

""" third-party """
try:
    from jsonschema import SchemaError, ValidationError, validate
except ImportError as e:
    print('Error: {}'.format(e))
    print('Try app.py --lib or adding jsonschema to setup.py')

# Load Schema
schema_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'tcex_json_schema.json')
with open(schema_file) as fh:
    schema = json.load(fh)

class TcExLocal:
    """
    Class to support running local instance TcEx Apps
    """

    def __init__(self):
        """
        """
        # Required Argument
        self._parsed = False  # only parse once from user
        self._parser = argparse.ArgumentParser()

        self._install_json = {}
        # self.load_install_json()
        self._app_packages = []

        self._required_arguments()
        self._args, unknown = self._parser.parse_known_args()

    @property
    def args(self):
        """The parsed args

        Returns:
            (namespace): ArgParser parsed arguments
        """
        if not self._parsed:
            self._args, unknown = self._parser.parse_known_args()
            self._parsed = True

        return self._args

    @property
    def parser(self):
        """The ArgParser parser object

        Returns:
            (ArgumentParser): TcEx local parser object
        """
        return self._parser

    @property
    def run(self):
        """Execute the script with arguments provided in tc.json."""

        if not os.path.isfile(self._args.script):
            msg = 'Provided script file does not exist ({0}).'
            msg.format(self._args.script)
            self._exit(msg, 1)

        command = '{0} . {1} {2}'.format(
            self._args.python,
            self._args.script.split('.')[0],
            self._parameters)
        os.system(command)

        msg = 'Executed: {0}'.format(command)
        self._exit(msg)

    @property
    def _config(self):
        """Configuration value retrieve from JSON configuration file

        Returns:
            (dict): Dictionary of configuration parameters
        """
        if not os.path.isfile(self._args.config):
            msg = 'Provided config file does not exist ({0}).'
            msg.format(self._args.config)
            self._exit(msg, 1)

        with open(self._args.config) as data_file:
            # return json.load(data_file)
            data = json.load(data_file)

        test_section = self._args.test
        if self._args.test is None:
            test_section = 'default'

        data = data.get(test_section)
        if data is None:
            msg = 'The tc.json configuration test named {} was not found.'.format(test_section)
            self._exit(msg, 1)

        return data

    def _exit(self, message, code=0):
        """Exit the script

        Args:
            message (str): An exit message
            code (Optional [int]): The exit status code
        """
        print(message)
        sys.exit(code)

    @property
    def _parameters(self):
        """Build CLI arguments to pass to script on the command line.

        This method takes the json data and covert it to CLI args for the execution
        of the script.

        Returns:
            (str): A string containing all parameter to pass to script
        """
        parameters = ' '
        for config_key, config_val in self._config.items():
            if isinstance(config_val, bool):
                if config_val:
                    parameters += '--{0} '.format(config_key)
            elif isinstance(config_val, list):
                for val in config_val:
                    parameters += '--{0} {1} '.format(
                        config_key, self._wrap(val))
            elif isinstance(config_val, (int, float, long, str, unicode)):
                parameters += '--{0} {1} '.format(
                    config_key, self._wrap(str(config_val)))
            else:
                msg = '{0} error: ({1}) is not of Type list or str. ({2})'
                msg.format(type(config_val), config_val)
                self._exit(msg, 1)

        return parameters

    def _required_arguments(self):
        """Required arguments for this class to function"""

        # actions
        self._parser.add_argument(
            '--lib', action='store_true', help='Gen the libs')
        self._parser.add_argument(
            '--package', action='store_true', help='Package the app')
        self._parser.add_argument(
            '--run', action='store_true', help='Run the app')
        self._parser.add_argument(
            '--validate', action='store_true', help='Validate JSON configs')

        # run args
        self._parser.add_argument(
            '--config', help='The configuration file', default='tc.json')
        self._parser.add_argument(
            '--python', help='The python executable', default='python')
        self._parser.add_argument(
            '--script', help='The script to be executed', required=False)
        self._parser.add_argument(
            '--test', help='The test to be executed', default=None)

        # validate args
        self._parser.add_argument(
            '--install_json', help='The install.json filename', default='install.json')

        # package args
        self._parser.add_argument(
            '--collection', help='Build app collection', action='store_true')
        self._parser.add_argument(
            '--zip_out', help='The zip output path', default=None)

    @staticmethod
    def _wrap(data):
        """Wrap any parameters that contain spaces

        Returns:
            (string): String containing parameters wrapped in double quotes
        """
        if len(re.findall(r'[!\-\s\$]{1,}', data)) > 0:
            data = '\'{}\''.format(data)
        return data

    ############################################################################
    # bcs - add comment and cleanup
    ############################################################################

    def load_install_json(self):
        """Read the install.json file"""
        with open('install.json') as fh:
            self._install_json = json.load(fh)

    def gen_lib(self):
        """Build libs locally for app

        Using the setup.py this method will install all required python modules locally
        to be used for local testing.
        """
        lib_directory = 'lib_{}.{}.{}'.format(
                sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
        app_path = os.getcwd()
        app_name = os.path.basename(app_path)

        lib_path = os.path.join(app_path, lib_directory)
        if not os.path.isdir(lib_path):
            os.mkdir(lib_path)

        os.environ['PYTHONPATH'] = '{0}'.format(lib_path)
        stdout = sys.stdout
        stderr = sys.stderr
        try:
            with open(os.path.join(app_path, '{}-libs.log'.format(app_name)), 'w') as log:
                sys.stdout = log
                sys.stderr = log
                easy_install.main(['-axZ', '-d', lib_path, str(app_path)])
        except SystemExit as e:
            raise Exception(str(e))
        finally:
            sys.stdout = stdout
            sys.stderr = stderr

        if len(os.listdir(lib_path)) == 0:
            raise Exception('Encountered error running easy_install for {}.  Check log file for details.'.format(app_name))

        build_path = os.path.join(app_path, 'build')
        if os.access(build_path, os.W_OK):
            shutil.rmtree(build_path)
        temp_path = os.path.join(app_path, 'temp')
        if os.access(temp_path, os.W_OK):
            shutil.rmtree(temp_path)

    def package(self):
        """Package the app for deployment

        This method will package the app for deployment to ThreatConnect.  It will download
        all required dependencies and include them in the package.  Validating of the
        install.json file or files will be automatically run before packaging the app.
        """

        # bcs - other options
        # app_name = self._install_json['displayName'].replace(' ', '_')
        # app_path = os.path.dirname(os.path.realpath(__file__))

        lib_directory = 'lib_{}.{}.{}'.format(
            sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
        app_path = os.getcwd()
        contents = os.listdir(app_path)

        # create build directory
        tmp_path = os.path.join(os.sep, 'tmp', 'tcex_builds')
        if not os.path.isdir(tmp_path):
            os.mkdir(tmp_path)

        # copy project directory to temp location to use as template for multiple builds
        base_name = os.path.basename(app_path)
        template_app_path = os.path.join(tmp_path, base_name)
        if os.access(template_app_path, os.W_OK):
            # cleanup any previous failed builds
            shutil.rmtree(template_app_path)

        # ignore unwanted files from build to ensure app packages are minimum size
        ignore_patterns = shutil.ignore_patterns(
            '*.git*', 'lib', '*log', '*python-version', 'tc.json', '*.tcx')
        shutil.copytree(app_path, template_app_path, False, ignore_patterns)

        for install_json in contents:
            if 'install.json' not in install_json:
                continue

            self.validate(install_json)

            base_name = os.path.basename(app_path)
            if install_json == 'install.json':
                app_name = base_name
            else:
                app_name = install_json.split('.')[0]

            tmp_app_path = os.path.join(tmp_path, app_name)
            if tmp_app_path != template_app_path:
                shutil.copytree(template_app_path, tmp_app_path)

            lib_path = os.path.join(tmp_app_path, lib_directory)
            if os.access(lib_path, os.W_OK):
                shutil.rmtree(lib_path)
            os.mkdir(lib_path)

            os.environ['PYTHONPATH'] = '{0}'.format(lib_path)
            stdout = sys.stdout
            stderr = sys.stderr
            try:
                with open(os.path.join(app_path, '{}-package.log'.format(app_name)), 'w') as log:
                    sys.stdout = log
                    sys.stderr = log
                    easy_install.main(['-axZ', '-d', lib_path, str(tmp_app_path)])
            except SystemExit as e:
                raise Exception(str(e))
            finally:
                sys.stdout = stdout
                sys.stderr = stderr

            if len(os.listdir(lib_path)) == 0:
                raise Exception('Encountered error running easy_install for {}.  Check log file for details.'.format(app_name))

            # cleanup
            git_path = os.path.join(tmp_app_path, '.git')
            if os.access(git_path, os.W_OK):
                shutil.rmtree(git_path)
            build_path = os.path.join(tmp_app_path, 'build')
            if os.access(build_path, os.W_OK):
                shutil.rmtree(build_path)

            # rename install.json
            if install_json != 'install.json':
                install_json_path = os.path.join(tmp_app_path, install_json)
                if os.access(build_path, os.W_OK):
                    shutil.rmtree(install_json_path)
                shutil.move(install_json_path, os.path.join(tmp_app_path, 'install.json'))

            # zip build directory
            zip_file = os.path.join(app_path, app_name)
            if self._args.zip_out is not None and os.access(self._args.zip_out, os.W_OK):
                zip_file = os.path.join(self._args.zip_out, app_name)

            zip_file_zip = '{}.zip'.format(zip_file)
            zip_file_tcx = '{}.tcx'.format(zip_file)
            shutil.make_archive(zip_file, 'zip', tmp_path, app_name)
            shutil.move(zip_file_zip, zip_file_tcx)
            self._app_packages.append(zip_file_tcx)

            # cleanup build directory
            if install_json != 'install.json':
                shutil.rmtree(tmp_app_path)

        if self._args.collection and len(self._app_packages) > 0:
            collection_file = '{}.zip'.format(base_name)
            z = zipfile.ZipFile(collection_file, 'w')
            for app in self._app_packages:
                z.write(app, os.path.basename(app))
            z.close
            if self._args.zip_out is not None and os.access(self._args.zip_out, os.W_OK):
                collection_zip = os.path.join(self._args.zip_out, collection_file)
                shutil.move(collection_file, collection_zip)

        # cleanup template directory
        if os.access(template_app_path, os.W_OK):
            shutil.rmtree(template_app_path)

    def validate(self, install_json):
        """Validate install.json file for required parameters"""

        # install.json validation
        try:
            with open(install_json) as fh:
                data = json.loads(fh.read())
            validate(data, schema)
            print('{} is valid'.format(install_json))
        except SchemaError as e:
            print('{} is invalid "{}"'.format(install_json, e))
        except ValidationError as e:
            print('{} is invalid "{}"'.format(install_json, e))



""" standard """
""" third-party """
import redis

""" custom """

class TcExRedis(object):
    """
    """
    def __init__(self, host, port, rhash):
        """ """
        self._hash = rhash
        # self._r = redis.StrictRedis(host=host, port=port, db=0)
        self._r = redis.StrictRedis(host=host, port=port)

    def create(self, key, value):
        """Create key/valu pair in Redis"""
        return self._r.hset(self._hash, key, value)

    def read(self, key):
        """Read data from Redis for the provided key"""
        return self._r.hget(self._hash, key)


import re
from setuptools import setup, find_packages

with open('tcex/__init__.py', 'r') as fd:
    version = re.search(
        r'^__version__(?:\s+)?=(?:\s+)?[\'|\"]((?:[0-9]{1,3}(?:\.)?){1,3})[\'|\"]', fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

setup(
    author='ThreatConnect (support@threatconnect.com)',
    author_email = 'support@threatconnect.com',
    description = 'ThreatConnect Exchange App Framework',
    download_url = 'https://github.com/ThreatConnect-Inc/tcex/tarball/{}'.format(version),
    install_requires = [
        'hiredis==0.2.0',
        'jsonschema==2.5.1',
        'python-dateutil==2.4.2',
        'redis==2.10.5',
        'requests==2.11.1'
    ],
    license = 'Apache License, Version 2',
    name = 'tcex',
    packages = find_packages(),
    package_data = {'': ['tcex_json_schema.json']},
    url = 'https://github.com/ThreatConnect-Inc/tcex',
    use_2to3 = True,
    # use_2to3_exclude_fixers = ['lib2to3.fixes.fix_print'],
    version = version
)